pub type TypeName = &'static str;

#[derive(Clone)]
pub struct Parameter {
    pub name: String,
    pub ty: TypeName,
}

#[derive(Clone)]
pub struct Method
{
    /// A pointer to the function.
    pub method_pointer: *mut fn(),

    /// The name of the method.
    pub name: &'static str,
    /// The parameter list.
    pub parameters: Vec<Parameter>,
    /// The return type (if any).
    pub ret: Option<TypeName>,

    /// Whether the method has a receiver.
    pub is_static: bool,
}

#[derive(Clone)]
pub struct Field
{
    /// The number of bytes from the start of the structure.
    pub field_offset: usize,

    pub ty: TypeName,
    pub name: &'static str,
}

#[derive(Clone)]
pub struct Class
{
    pub name: String,
    pub fields: Vec<Field>,
    pub methods: Vec<Method>,
}

pub trait PluggableFields
{
    fn pluggable_fields(&self) -> Vec<Field>;
}

pub trait PluggableMethods
{
    fn pluggable_methods(&self) -> Vec<Method>;
}

/// An object that can be plugged into a scripting language.
///
/// Can be automatically derived by placing `#[pluggable]` on a struct.
///
/// When implementing `Pluggable`, it is necessary to have both a `struct` and an
/// `impl`, with both having the `#[pluggable]` attribute.
///
/// This has been split up into two other traits
///
/// * `PluggableFields`
/// * `PluggableMethods
///
/// This is because the syntax extender can not see all of the code -
/// it can only see the thing it's operating on. This is why we need
/// an attribute on the `struct` and the `impl`. We can't implement
/// a trait twice, each time overriding a specific method.
///
/// Because of this, it is necessary to mark both the `struct` and the `impl`
/// with the `pluggable` attribute. It's not the prettiest, but it does work.
pub trait Pluggable : PluggableFields + PluggableMethods
{
    fn name(&self) -> &'static str;

    fn fields(&self) -> Vec<Field> { PluggableFields::pluggable_fields(self) }
    fn methods(&self) -> Vec<Method> { PluggableMethods::pluggable_methods(self) }

    fn class(&self) -> Class {
        Class {
            name: self.name().to_owned(),
            fields: self.fields(),
            methods: self.methods(),
        }
    }
}

